"""Hierarchical Clustering estimators."""

# Author: Hugo Delatte <delatte.hugo@gmail.com>
# SPDX-License-Identifier: BSD-3-Clause

from skfolio.cluster._hierarchical import HierarchicalClustering, LinkageMethod

__all__ = ["HierarchicalClustering", "LinkageMethod"]
